// Queue Class Implementation
class Queue {
    constructor(maxSize = Infinity) {
        this.items = []; // Array to store queue elements
        this.maxSize = maxSize; // Maximum size of the queue
    }

    // Enqueue: Add an element to the end of the queue
    enqueue(element) {
        if (this.isFull()) {
            console.log("Queue is full! Cannot enqueue.");
            return;
        }
        this.items.push(element);
        console.log(`${element} enqueued to the queue`);
    }

    // Dequeue: Remove an element from the front of the queue
    dequeue() {
        if (this.isEmpty()) {
            console.log("Queue is empty! Cannot dequeue.");
            return null;
        }
        const removed = this.items.shift(); // Remove first element
        console.log(`${removed} dequeued from the queue`);
        return removed;
    }

    // Peek: Get the front element of the queue without removing it
    peek() {
        if (this.isEmpty()) {
            console.log("Queue is empty! Nothing to peek.");
            return null;
        }
        const front = this.items[0];
        console.log(`Front of the queue is: ${front}`);
        return front;
    }

    // Get the rear element of the queue
    rear() {
        if (this.isEmpty()) {
            console.log("Queue is empty! Nothing at the rear.");
            return null;
        }
        const rear = this.items[this.items.length - 1];
        console.log(`Rear of the queue is: ${rear}`);
        return rear;
    }

    // isEmpty: Check if the queue is empty
    isEmpty() {
        return this.items.length === 0;
    }

    // isFull: Check if the queue has reached its maximum size
    isFull() {
        return this.items.length >= this.maxSize;
    }

    // Size: Get the number of elements in the queue
    size() {
        const queueSize = this.items.length;
        console.log(`Current queue size: ${queueSize}`);
        return queueSize;
    }

    // Traverse: Print all elements in the queue
    traverse() {
        if (this.isEmpty()) {
            console.log("Queue is empty! Nothing to traverse.");
            return;
        }
        console.log("Queue elements:");
        this.items.forEach((element, index) => {
            console.log(`Index ${index}: ${element}`);
        });
    }

    // Display queue as a formatted string
    toString() {
        const queueString = this.items.join(" <- ");
        console.log(`Queue as string: ${queueString}`);
        return queueString;
    }

    // Contains: Check if the queue contains a specific element
    contains(element) {
        const exists = this.items.includes(element);
        console.log(`Queue contains ${element}: ${exists}`);
        return exists;
    }

    // Copy a subset of the queue
    copySubset(startIndex, endIndex) {
        if (startIndex < 0 || endIndex > this.items.length || startIndex >= endIndex) {
            console.log("Invalid range for subset copy.");
            return null;
        }
        const subset = new Queue();
        subset.items = this.items.slice(startIndex, endIndex);
        console.log(`Subset copied from index ${startIndex} to ${endIndex - 1}`);
        return subset;
    }

    // Clear: Remove all elements from the queue
    clear() {
        this.items = [];
        console.log("Queue cleared!");
    }

    // Reverse the queue
    reverse() {
        if (this.isEmpty()) {
            console.log("Queue is empty! Cannot reverse.");
            return;
        }
        this.items.reverse();
        console.log("Queue reversed!");
    }

    // Search for an element in the queue
    search(element) {
        const index = this.items.indexOf(element);
        if (index !== -1) {
            console.log(`${element} found at index ${index} in the queue.`);
            return index;
        } else {
            console.log(`${element} not found in the queue.`);
            return -1;
        }
    }

    // Clone the queue
    clone() {
        const clonedQueue = new Queue();
        clonedQueue.items = [...this.items];
        console.log("Queue cloned!");
        return clonedQueue;
    }

    // Merge another queue into this queue
    merge(anotherQueue) {
        if (!(anotherQueue instanceof Queue)) {
            console.log("Can only merge with another queue.");
            return;
        }
        this.items = [...this.items, ...anotherQueue.items];
        console.log("Queues merged!");
    }
}

// Example Usage
const queue = new Queue(10);

console.log("=== Enqueue Operations ===");
queue.enqueue(10);
queue.enqueue(20);
queue.enqueue(30);
queue.enqueue(40);
queue.enqueue(50);

console.log("\n=== Traverse Queue ===");
queue.traverse();

console.log("\n=== Peek and Rear Operations ===");
queue.peek();
queue.rear();

console.log("\n=== Dequeue Operations ===");
queue.dequeue();
queue.dequeue();

console.log("\n=== Traverse Queue After Dequeue ===");
queue.traverse();

console.log("\n=== Size, Full, and Empty Checks ===");
queue.size();
console.log(`Is the queue empty? ${queue.isEmpty()}`);
console.log(`Is the queue full? ${queue.isFull()}`);

console.log("\n=== Reverse Queue ===");
queue.reverse();
queue.traverse();

console.log("\n=== Search and Contains Operations ===");
queue.search(30); // Existing element
queue.search(100); // Non-existing element
queue.contains(30);
queue.contains(100);

console.log("\n=== Subset Copy ===");
const subsetQueue = queue.copySubset(1, 3);
subsetQueue.traverse();

console.log("\n=== Clone Queue ===");
const clonedQueue = queue.clone();
console.log("Cloned Queue:");
clonedQueue.traverse();

console.log("\n=== Merge Queues ===");
const anotherQueue = new Queue();
anotherQueue.enqueue(60);
anotherQueue.enqueue(70);
anotherQueue.enqueue(80);

console.log("Another Queue:");
anotherQueue.traverse();

console.log("Merging queues...");
queue.merge(anotherQueue);
queue.traverse();

console.log("\n=== Display Queue as String ===");
queue.toString();

console.log("\n=== Clear the Queue ===");
queue.clear();

console.log("\n=== Final Check ===");
queue.size();
console.log(`Is the queue empty? ${queue.isEmpty()}`);
